
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#ifndef PG_OBJECT_IDENTIFIER_CLASS_H
#define PG_OBJECT_IDENTIFIER_CLASS_H

#include "base/UniqId.h"

// a set of attributes that uniquely identifies a Polygraph object
class ObjId {
	protected:
		typedef enum { flgNone = 0, flgRepeat = 1, flgForeign = 2,
			flgCachable = 4, flgHit = 8, flgIms = 16, flgReload = 32,
			flgHot = 64, flgRediredReq = 128, flgRepToRedir = 256 } Flag;

	public:
		ObjId() { reset(); }

		void reset();

		const UniqId &world() const { return theWorld; }
		int viserv() const { return theViserv; }
		int target() const { return theTarget; }
		int type() const { return theType; }
		int name() const { return theName; }

		bool direct() const;
		bool embedded() const;
		bool repeat() const { return (theFlags & flgRepeat) != 0; }
		bool hot() const { return (theFlags & flgHot) != 0; }
		bool foreign() const { return (theFlags & flgForeign) != 0; }

		bool basic() const { return !(ims() || reload() || rediredReq() || repToRedir()); }
		bool ims() const { return (theFlags & flgIms) != 0; }
		bool reload() const { return (theFlags &  flgReload) != 0; }
		bool rediredReq() const { return (theFlags &  flgRediredReq) != 0; }
		bool repToRedir() const { return (theFlags &  flgRepToRedir) != 0; }
		
		bool offeredHit() const { return basic() && repeat() && cachable(); }
		bool hit() const { return (theFlags & flgHit) != 0; }
		bool cachable() const { return (theFlags & flgCachable) != 0; }
		bool fill() const;

		void repeat(bool be) { setFlag(flgRepeat, be); }
		void hot(bool be) { setFlag(flgHot, be); }
		void foreign(bool be) { setFlag(flgForeign, be); }

		void ims(bool be) { setFlag(flgIms, be); }
		void reload(bool be) { setFlag(flgReload, be); }
		void rediredReq(bool be) { setFlag(flgRediredReq, be); }
		void repToRedir(bool be) { setFlag(flgRepToRedir, be); }
		
		void hit(bool be) { setFlag(flgHit, be); }
		void cachable(bool be) { setFlag(flgCachable, be); }

		operator void*() const { return theName > 0 ? (void*)-1 : 0; }

		void world(const UniqId &aWorld) { theWorld = aWorld; }
		void viserv(int aViserv) { theViserv = aViserv; }
		void target(int aTarget) { theTarget = aTarget; }
		void type(int aType) { theType = aType; }
		void name(int aName) { theName = aName; }

		// XXX: for parsing
		UniqId &world() { return theWorld; }
		int &name() { return theName; }
		int &type() { return theType; }

		int hash() const;

		inline bool operator <(const ObjId &id) const;  // arbitrary order
		inline bool operator ==(const ObjId &id) const;
		inline bool operator !=(const ObjId &id) const;

		ostream &printFlags(ostream &os) const;

		void store(OLog &ol) const;
		void load(ILog &il);

	protected:
		inline void setFlag(int flag, bool set);

	protected:
		UniqId theWorld;   // world id
		int theViserv;     // visible server (index to Viservs)
		int theType;       // object content type
		int theName;       // object "name"

		int theTarget;     // actual host (index to HostMap)

		unsigned theFlags;
};


inline
bool ObjId::operator <(const ObjId &id) const {
	if (theName < id.theName)
		return true;
	if (theName > id.theName)
		return false;

	if (theViserv < id.theViserv)
		return true;
	if (theViserv > id.theViserv)
		return false;
	
	return theWorld < id.theWorld;
}

inline
bool ObjId::operator ==(const ObjId &id) const {
	return 
		theName == id.theName &&
		theViserv == id.theViserv &&
		theWorld == id.theWorld;
}

inline
bool ObjId::operator !=(const ObjId &id) const {
	return !(*this == id);
}

inline
void ObjId::setFlag(int flag, bool set) {
	if (set)
		theFlags |= flag;
	else
		theFlags &= ~flag;
}

inline
OLog &operator <<(OLog &ol, const ObjId &r) {
	r.store(ol);
	return ol;
}

inline
ILog &operator >>(ILog &il, ObjId &r) {
	r.load(il);
	return il;
}

#endif
