
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/h/math.h"

#include "xstd/TblDistr.h"
#include "pgl/PglIntSym.h"
#include "pgl/PglArraySym.h"

#include "xstd/rndDistrs.h"
#include "xstd/gadgets.h"


ArraySym::ArraySym(const String &anItemType): 
	ListSym(anItemType ? (const char *)(anItemType + "[]") : "Array"),
	theItemType(anItemType), warnedBadProbs(false) {
}

bool ArraySym::isA(const String &type) const {
	return ListSym::isA(type) || type == "Array";
}

SynSym *ArraySym::dupe(const String &type) const {
	String h = type;
	String itemType;
	if (const char *p = h.str("[]"))
		itemType = h(0, p - h);
	else
	if (isA(type))
		itemType = theItemType;
	else
		return 0;

	ArraySym *clone = cloneType(itemType);

	if (!clone->absorb(*this)) {
		delete clone;
		return 0;
	}

	clone->syncProbs();
	for (int i = 0; i < clone->count(); ++i)
		clone->prob(i, theProbs[i]);

	return clone;
}

ArraySym *ArraySym::cloneType(const String &itemType) const {
	return new ArraySym(itemType);
}

void ArraySym::addProb(const SynSym &s, double p) {
	ListSym::add(s);
	prob(count()-1, p);
}

void ArraySym::addFailure(const SynSym &s) {
	cerr << s.loc() << "cannot add `" << s.type() << "' item to `"
		<< (theItemType ? (const char*) type() : "yet-type-less")
		<< "' array" << endl << xexit;
}

bool ArraySym::cadd(const SynSym &s) {
	// absorb other arrays (mimic Perl list behavior)
	if (s.isA("Array"))
		return absorb((ArraySym&)s.cast("Array"));

	// note: anonymous array has a type of its first item
	const String it = theItemType ? theItemType : s.type();

	if (SynSym *sclone = s.clone(it)) {

		if (!theItemType) {
			theItemType = it;
			type(it + "[]");
		}

		doAdd(sclone);
		syncProbs();
		return true;
	}

	return false;
}

ostream &ArraySym::print(ostream &os, const String &pfx) const {
	Array<double> probs;
	makeProbs(probs);
	const bool printProbs = true; // theUndefProbCnt != count();
	os << '[';
	for (int i = 0; i < count(); ++i) {
		if (i) os << ',';
		Assert(item(i));
		item(i)->print(os, pfx);
		Assert(probs[i] >= 0);
		if (printProbs)
			os << ':' << (probs[i]*100) << '%';
	}
	return os << ']';
}

void ArraySym::prob(int idx, double p) {
	syncProbs();
	Assert(idx < theProbs.count());
	theProbs[idx] = p;
}

// count the number of undefiend probs and
// get the sum of defined probs
double ArraySym::statProbs() const {
	Assert(theProbs.count() == count()); // synced

	theUndefProbCnt = 0;
	double sum = 0;
	for (int i = 0; i < theProbs.count(); ++i) {
		if (theProbs[i] > 0)
			sum += theProbs[i];
		else
		if (theProbs[i] < 0)
			theUndefProbCnt++;
	}
	return sum;
}

void ArraySym::copyProbs(Array<double> &res) const {
	checkProbs();
	makeProbs(res);
}

void ArraySym::makeProbs(Array<double> &res) const {
	const double sum = statProbs();

	if (sum >= 1.01)
		cerr << loc() << "array probabilities add up to " 
			<< 100*sum << "% which is more than 100%" << endl << xexit;

	// copy values
	res.stretch(theProbs.count());
	res.count(theProbs.count());
	for (int i = 0; i < theProbs.count(); ++i)
		res[i] = theProbs[i];

	// assign sum leftovers to the undefined probs
	if (theUndefProbCnt) {
		const double restp = Max((1.0-sum)/theUndefProbCnt, 0.0);
		for (int i = 0; i < res.count(); ++i) {
			if (res[i] < 0)
				res[i] = restp;
		}
	}
}

void ArraySym::makeProbs() {
	checkProbs();
	makeProbs(theProbs);
}

void ArraySym::checkProbs() const {
	const double sum = statProbs();
	if (sum <= 0.99 && !theUndefProbCnt && !warnedBadProbs) {
		print(cerr << loc()) << endl;
		cerr << loc() << "array probabilities (above) add up to " << 100*sum << "% (not 100%) and no wild-card entries are present;"
			<< " missing percents will be spread out proportionally among array entries" << endl;
		warnedBadProbs = true;
	}
}

// ensure that all probabilities are off by default
void ArraySym::syncProbs() {
	while (theProbs.count() < count())
		theProbs.append(-1);
}

bool ArraySym::probsVary() const {
	checkProbs();
	if (theUndefProbCnt == theProbs.count())
		return false;

	Array<double> probs;
	makeProbs(probs);

	if (!probs.count())
		return false;

	double sample = probs.last();
	for (int i = 0; i < probs.count(); ++i) {
		if (Abs(probs[i] - sample)*1e6 >= 1)
			return true;
	}

	return false;
}

// uses a table distribution to model item probabilities
// selector is a distribution of item positions
// XXX: if no probs specified, we use uniform distribution to save memory
//      ideally, table distr should be smart enough to detect this case
RndDistr *ArraySym::makeSelector(const String &name) {
	makeProbs();

	if (theUndefProbCnt == theProbs.count())
		return new UnifDistr(new RndGen, 0, theProbs.count());

	return TblDistr::FromDistrTable(name, theProbs);
}
