
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "xstd/Rnd.h"
#include "base/ObjId.h"
#include "base/OLog.h"
#include "base/ILog.h"
#include "runtime/PubWorld.h"


Array<PubWorld*> ThePubWorlds;


/* PubWorldSlice */

void PubWorldSlice::update(const ObjWorld &w) {
	// try to preserve consistency: update all or nothing
	if (w.newer(*this)) {
		size(w.size());
		wss(w.wss());
		hotPos(w.hotPos());
	}
}


/* PubWorld */

int PubWorld::PositiveWss() {
	int sum = -1;
	for (int i = 0; i < ThePubWorlds.count(); ++i) {
		const int w = ThePubWorlds[i]->positiveWss();
		if (w > 0) {
			if (sum < 0)
				sum = 0;
			sum += w;
		}
	}
	return sum;
}


PubWorld::PubWorld(): theLocalSlice(-1), theNextSliceToSync(0) {
}

PubWorld::PubWorld(const UniqId &localId): theLocalSlice(0), theNextSliceToSync(0) {
	Assert(localId);
	PubWorldSlice local;
	local.id(localId);
	theSlices.append(local);
}

bool PubWorld::find(const UniqId &id, int &sliceIdx) const {
	// XXX: slow
	for (sliceIdx = 0; sliceIdx < theSlices.count(); ++sliceIdx) {
		if (theSlices[sliceIdx].id() == id)
			return true;
	}
	return false;
}

const PubWorldSlice *PubWorld::sliceToSync() const {
	if (!theSlices.count())
		return 0;

	theNextSliceToSync %= theSlices.count();
	return &theSlices[theNextSliceToSync++];
}

const PubWorldSlice *PubWorld::newerSlice(const ObjWorld &oldSlice, int idx) const {
	for (int count = theSlices.count(); count--; ++idx) {
		idx %= theSlices.count();
		const PubWorldSlice &slice = theSlices[idx];
		if (slice) {
			if (slice.id() == oldSlice.id()) {
				if (slice.newer(oldSlice)) // known, newer slice
					return &slice;
			} else
				return &slice;  // unknown (other) slice, possibly new
		}
	}
	// there is only one slice and it is not newer
	return 0;
}

PubWorldSlice &PubWorld::getSlice(RndGen &rng, SliceKind kind) {
	int count = theSlices.count();
	for (int idx = rng(0, count); count--; ++idx) {
		idx %= theSlices.count();
		PubWorldSlice &slice = theSlices[idx];
		if ((slice.*kind)())
			return slice;
	}

	Assert(false); // getSlice must be called if can"Kind"() only
	return localSlice();
}

// fails if there is no local slice (e.g., server side)
PubWorldSlice &PubWorld::localSlice() {
	return sliceAt(theLocalSlice);
}

const PubWorldSlice &PubWorld::localSlice() const {
	return sliceAt(theLocalSlice);
}

const PubWorldSlice &PubWorld::sliceAt(int idx) const {
	Assert(0 <= idx && idx < theSlices.count());
	return theSlices[idx];
}

PubWorldSlice &PubWorld::sliceAt(int idx) {
	Assert(0 <= idx && idx < theSlices.count());
	return theSlices[idx];
}

// sums positive wss()s
int PubWorld::positiveWss() const {
	int wss = -1;
	for (int i = 0; i < theSlices.count(); ++i) {
		if (theSlices[i].wss() > 0) {
			if (wss < 0)
				wss = 0;
			wss += theSlices[i].wss();
		}
	}
	return wss;
}

int PubWorld::size() const {
	int sum = 0;
	for (int i = 0; i < theSlices.count(); ++i)
		sum += theSlices[i].size();
	return sum;
}

// XXX: these kind of loops are slow
bool PubWorld::canRepeat() const {
	for (int i = 0; i < theSlices.count(); ++i) {
		if (theSlices[i].canRepeat())
			return true;
	}
	return false;
}

bool PubWorld::canProduce() const {
	// to avoid race conditions (two robots producing a false miss), 
	// only local slices can produce 
	return theLocalSlice >= 0 && localSlice().canProduce();
}

void PubWorld::repeat(ObjId &oid, ObjSelector *sel, RndGen &rng) {
	PubWorldSlice &slice = getSlice(rng, &PubWorldSlice::canRepeat);
	slice.repeat(oid, sel, rng);
}

void PubWorld::produce(ObjId &oid, RndGen &rng) {
	Assert(theLocalSlice >= 0);
	localSlice().produce(oid, rng);
}

void PubWorld::freezeWss() {
	localSlice().freezeWss();
}

void PubWorld::add(const PubWorldSlice &slice) {
	theSlices.append(slice);
	//if (slice.wss() > 0)
	//	ObjWorld::TheWss += slice.wss(); // XXX: CHECK: (was enabled) used anywhere?
}

void PubWorld::store(OLog &ol) const {
	// store local information only
	localSlice().store(ol); // XXX: would not this fail on server side?
}

